from typing import Callable
from dataclasses import dataclass
from busline.event.event import Event
from busline.client.subscriber.event_handler.event_handler import EventHandler


@dataclass
class ClosureEventHandler(EventHandler):
    """
    Event handler which use a pre-defined callback as `on_event`

    Author: Nicola Ricciardi
    """

    on_event_callback: Callable[[str, Event], None]

    async def handle(self, topic: str, event: Event):
        self.on_event_callback(topic, event)



import logging
from dataclasses import dataclass, field
from typing import Optional, override

from busline.client.subscriber.topic_subscriber import TopicSubscriber
from busline.local.eventbus.eventbus import EventBus
from busline.exceptions import EventBusClientNotConnected


@dataclass
class LocalEventBusSubscriber(TopicSubscriber):
    """
    Subscriber topic-based which works with local eventbus

    Author: Nicola Ricciardi
    """

    eventbus: EventBus
    connected: bool = field(default=False)

    @override
    async def connect(self):
        logging.info(f"subscriber {self.identifier} connecting...")
        self.connected = True

    @override
    async def disconnect(self):
        logging.info(f"subscriber {self.identifier} disconnecting...")
        self.connected = False

    @override
    async def _internal_subscribe(self, topic: str, **kwargs):
        if not self.connected:
            raise EventBusClientNotConnected()

        self.eventbus.add_subscriber(topic, self)

    @override
    async def _internal_unsubscribe(self, topic: Optional[str] = None, **kwargs):
        if not self.connected:
            raise EventBusClientNotConnected()

        self.eventbus.remove_subscriber(self, topic)



from dataclasses import dataclass, field
from typing import Dict, Type
from functools import wraps
from busline.event.event import Event, EventPayload
from busline.utils.singleton import Singleton


class EventRegistry(metaclass=Singleton):
    """
    Registry to manage different event types

    Author: Nicola Ricciardi
    """

    __associations: Dict[str, Type[EventPayload]] = {}

    @property
    def associations(self) -> Dict[str, Type[EventPayload]]:
        return self.__associations

    def remove(self, event_type: str):
        """
        Remove an event type association
        """

        self.__associations.pop(event_type)

    def add(self, event_type: str, event_payload_class: Type[EventPayload]):
        """
        Add a new association between an event type and an event class
        """

        self.__associations[event_type] = event_payload_class

    def retrieve_class(self, event_type: str) -> Type[EventPayload]:
        """
        Retrieve event class of event input based on saved associations and given event type

        KeyError is raised if no association is found
        """

        return self.__associations[event_type]

    def build(self, event_type: str, format_type: str, serialized_payload: bytes) -> EventPayload:
        """
        Build EventPayload from Event, raises a KeyError if there is not the association in registry
        """

        event_class: Type[EventPayload] = self.retrieve_class(event_type)

        return event_class.deserialize(format_type, serialized_payload)



def registry(cls: Type[EventPayload]):

    event_type: str = cls.__name__

    # add event payload to registry
    reg = EventRegistry()
    reg.add(event_type, cls)

    return cls








from abc import ABC, abstractmethod
from uuid import uuid4
from typing import Optional
from dataclasses import dataclass, field


@dataclass(kw_only=True)
class EventBusConnector(ABC):
    """
    Abstract class which provides methods to interact with eventbus

    Author: Nicola Ricciardi
    """

    identifier: str = field(default=str(uuid4()))


    @abstractmethod
    async def connect(self):
        """
        Connect to eventbus

        :return:
        """

    @abstractmethod
    async def disconnect(self):
        """
        Disconnect to eventbus

        :return:
        """


import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from busline.event.event import Event
from busline.client.eventbus_connector import EventBusConnector


@dataclass
class Publisher(EventBusConnector, ABC):
    """
    Abstract class which can be implemented by your components which must be able to publish on eventbus

    Author: Nicola Ricciardi
    """

    def __repr__(self) -> str:
        return f"Publisher({self.identifier})"

    @abstractmethod
    async def _internal_publish(self, topic: str, event: Event, **kwargs):
        """
        Actual publish on topic the event

        :param topic:
        :param event:
        :return:
        """

    async def publish(self, topic: str, event: Event, **kwargs):
        """
        Publish on topic the event

        :param topic:
        :param event:
        :return:
        """

        logging.info(f"{self}: publish on {topic} -> {event}")
        await self.on_publishing(topic, event, **kwargs)
        await self._internal_publish(topic, event, **kwargs)
        await self.on_published(topic, event, **kwargs)


    async def on_publishing(self, topic: str, event: Event, **kwargs):
        """
        Callback called on publishing start

        :param topic:
        :param event:
        :return:
        """

    async def on_published(self, topic: str, event: Event, **kwargs):
        """
        Callback called on publishing end

        :param topic:
        :param event:
        :return:
        """


import asyncio
import logging
from abc import ABC
from typing import Dict, List, Callable, Optional, override
from dataclasses import dataclass, field
from busline.client.subscriber.event_handler.event_handler import EventHandler
from busline.client.subscriber.subscriber import Subscriber
from busline.event.event import Event
from busline.exceptions import EventHandlerNotFound


@dataclass(kw_only=True)
class TopicSubscriber(Subscriber, ABC):
    """
    Handles different topic events using ad hoc handlers defined by user,
    else it uses fallback handler if provided (otherwise throws an exception)

    Attributes:
        fallback_event_handler: event handler used for a topic if no event handler is specified for that topic
        handlers: event handler for each topic (i.e. key); notice that key can also be a string with wildcards
        topic_names_matcher: function used to check match between two topic name (with wildcards); default "t1 == t2"
        event_handler_always_required: raise an exception if no handlers are found for a topic

    Author: Nicola Ricciardi
    """

    fallback_event_handler: Optional[EventHandler] = field(default=None)
    handlers: Dict[str, EventHandler] = field(default_factory=dict)
    topic_names_matcher: Callable[[str, str], bool] = field(repr=False, default=lambda t1, t2: t1 == t2)
    event_handler_always_required: bool = field(default=False)

    @override
    async def _on_subscribing(self, topic: str, handler: Optional[EventHandler] = None, **kwargs):

        if self.fallback_event_handler is None:
            if self.event_handler_always_required:
                raise EventHandlerNotFound()
            else:
                logging.warning(f"{self}: event handler for topic '{topic}' not found")

    @override
    async def _on_subscribed(self, topic: str, handler: Optional[EventHandler] = None, **kwargs):

        self.handlers[topic] = handler

    @override
    async def _on_unsubscribed(self, topic: Optional[str], **kwargs):

        if topic is None:
            self.handlers = {}
        else:
            if topic in self.handlers:
                del self.handlers[topic]
            else:
                logging.warning(f"{self}: unsubscribed from unknown topic: {topic}")

    def __get_handlers_of_topic(self, topic: str) -> List[EventHandler]:

        handlers = []
        for t, h in self.handlers.items():
            if self.topic_names_matcher(topic, t):
                if h is not None:
                    handlers.append(h)
                else:
                    if self.fallback_event_handler is not None:
                        handlers.append(self.fallback_event_handler)
                    else:
                        if self.event_handler_always_required:
                            raise EventHandlerNotFound()
                        else:
                            logging.warning(f"{self}: event handler for topic '{topic}' not found")

        return handlers

    @override
    async def on_event(self, topic: str, event: Event):

        handlers_of_topic: List[EventHandler] = self.__get_handlers_of_topic(topic)

        if len(handlers_of_topic) > 0:
            tasks = [handler.handle(topic, event) for handler in handlers_of_topic]
            await asyncio.gather(*tasks)



from abc import ABC, abstractmethod
from typing import Optional

from busline.event.event import Event


class EventBusClient(ABC):

    @abstractmethod
    async def connect(self):
        pass

    @abstractmethod
    async def disconnect(self):
        pass

    @abstractmethod
    async def publish(self, topic: str, event: Event, **kwargs):
        pass

    @abstractmethod
    async def subscribe(self, topic: str, **kwargs):
        pass

    @abstractmethod
    async def unsubscribe(self, topic: Optional[str] = None, **kwargs):
        pass



import json
import unittest
from typing import Type, Self, Tuple, override
from dataclasses import dataclass

from busline.event.avro_payload import AvroEventPayload
from busline.event.event import Event, EventPayload
from busline.event.json_payload import JsonEventPayload
from busline.event.registry import EventRegistry, registry


@dataclass(frozen=True)
class EventPayload1(EventPayload):

    value: int

    def my_value1(self) -> int:
        return self.value

    def serialize(self) -> Tuple[str, bytes]:
        return "raw", self.value.to_bytes(1)

    @classmethod
    def deserialize(cls, payload_type: str, payload: bytes) -> Self:
        return EventPayload1(int.from_bytes(payload))

@dataclass(frozen=True)
@registry
class EventPayload2(EventPayload):

    value: int

    def my_value2(self) -> int:
        return self.value

    def serialize(self) -> Tuple[str, bytes]:
        return "raw", self.value.to_bytes(1)

    @classmethod
    def deserialize(cls, payload_type: str, payload: bytes) -> Self:
        return EventPayload2(int.from_bytes(payload))

@dataclass(frozen=True)
class MockUserCreationJsonPayload(JsonEventPayload):

    email: str
    password: str

    @classmethod
    @override
    def from_json(cls, json_str: str) -> Self:
        data = json.loads(json_str)

        return MockUserCreationJsonPayload(data["email"], data["password"])


@dataclass(frozen=True)
class MockUserCreationAvroPayload(AvroEventPayload):
    email: str
    password: str


class TestEventRegistry(unittest.TestCase):

    def test_raw_payload(self):
        event = Event(payload=EventPayload1(1))

        _, serialized_payload = event.payload.serialize()
        self.assertIs(type(serialized_payload), bytes)


    def test_json_payload(self):
        event = Event(payload=MockUserCreationJsonPayload("email", "password"))

        format_type, serialized_payload = event.payload.serialize()

        self.assertIs(type(serialized_payload), bytes)
        self.assertEqual(format_type, "application/json")

        payload: MockUserCreationJsonPayload = MockUserCreationJsonPayload.deserialize(format_type, serialized_payload)

        self.assertEqual(payload.email, event.payload.email)
        self.assertEqual(payload.password, event.payload.password)

    def test_registry(self):

        event_registry = EventRegistry()    # singleton

        event_registry.add("event_payload1", EventPayload1)

        self.assertEqual(len(event_registry.associations), 2)

        event_registry = EventRegistry()  # singleton

        generic_event1 = Event(payload=EventPayload1(1), event_type="event_payload1")
        generic_unknown_event = Event(payload=None, event_type="unknown")

        self.assertNotEqual(generic_event1.identifier, generic_unknown_event.identifier)

        format_type, serialized_payload = generic_event1.payload.serialize()
        event_payload1: EventPayload1 = event_registry.build(generic_event1.event_type, format_type, serialized_payload)

        self.assertEqual(event_payload1.value, 1)
        self.assertEqual(event_payload1.my_value1(), 1)

        self.assertRaises(KeyError, lambda: event_registry.retrieve_class(generic_unknown_event.event_type))

    def test_avro_payload(self):

        payload = MockUserCreationAvroPayload("email", "password")
        event = payload.into_event()

        format_type, serialized_payload = payload.serialize()

        deserialized_payload = MockUserCreationAvroPayload.deserialize(format_type, serialized_payload)

        self.assertEqual(payload.email, deserialized_payload.email)
        self.assertEqual(payload.password, deserialized_payload.password)


from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, Optional
from busline.exceptions import TopicNotFound
from busline.client.subscriber.subscriber import Subscriber
from busline.event.event import Event


@dataclass
class EventBus(ABC):
    """
    Abstract class used as base for new eventbus implemented in local projects.

    Author: Nicola Ricciardi
    """

    subscriptions: Dict[str, List[Subscriber]] = field(default_factory=dict)

    def __post_init__(self):

        self.reset_subscriptions()

    def reset_subscriptions(self):
        self.subscriptions = {}

    @property
    def topics(self) -> List[str]:
        return list(self.subscriptions.keys())

    def add_subscriber(self, topic: str, subscriber: Subscriber):
        """
        Add subscriber to topic

        :param topic:
        :param subscriber:
        :return:
        """

        self.subscriptions.setdefault(topic, [])
        self.subscriptions[topic].append(subscriber)

    def remove_subscriber(self, subscriber: Subscriber, topic: Optional[str] = None, raise_if_topic_missed: bool = False):
        """
        Remove subscriber from topic selected or from all if topic is None

        :param raise_if_topic_missed:
        :param subscriber:
        :param topic:
        :return:
        """

        if raise_if_topic_missed and topic is not None and topic not in self.subscriptions.keys():
            raise TopicNotFound(f"topic '{topic}' not found")

        for name in self.subscriptions.keys():

            if topic is None or self._topic_names_match(topic, name):
                if subscriber in self.subscriptions[name]:
                    self.subscriptions[name].remove(subscriber)


    def _topic_names_match(self, t1: str, t2: str):
        return t1 == t2

    def _get_topic_subscriptions(self, topic: str) -> List[Subscriber]:

        topic_subscriptions: List[Subscriber] = []
        for t, subs in self.subscriptions.items():
            if self._topic_names_match(t, topic):
                topic_subscriptions.extend(subs)

        return topic_subscriptions

    @abstractmethod
    async def put_event(self, topic: str, event: Event):
        """
        Put a new event in the bus and notify subscribers of corresponding
        event's topic

        :param topic:
        :param event:
        :return:
        """

        raise NotImplemented()




import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from busline.client.eventbus_connector import EventBusConnector
from busline.event.event import Event


@dataclass
class Subscriber(EventBusConnector, ABC):
    """
    Abstract class which can be implemented by your components which must be able to subscribe on eventbus

    Author: Nicola Ricciardi
    """

    def __repr__(self) -> str:
        return f"Subscriber({self.identifier})"

    @abstractmethod
    async def on_event(self, topic: str, event: Event):
        """
        Callback called when an event arrives from a topic
        """

    async def notify(self, topic: str, event: Event, **kwargs):
        """
        Notify subscriber
        """

        logging.info(f"{self}: incoming event on {topic} -> {event}")
        await self.on_event(topic, event)

    @abstractmethod
    async def _internal_subscribe(self, topic: str, **kwargs):
        """
        Actual subscribe to topic

        :param topic:
        :return:
        """

    @abstractmethod
    async def _internal_unsubscribe(self, topic: Optional[str] = None, **kwargs):
        """
        Actual unsubscribe to topic

        :param topic:
        :return:
        """

    async def subscribe(self, topic: str, **kwargs):
        """
        Subscribe to topic

        :param topic:
        :return:
        """

        logging.info(f"{self}: subscribe on topic {topic}")
        await self._on_subscribing(topic, **kwargs)
        await self._internal_subscribe(topic, **kwargs)
        await self._on_subscribed(topic, **kwargs)

    async def unsubscribe(self, topic: Optional[str] = None, **kwargs):
        """
        Unsubscribe to topic

        :param topic:
        :return:
        """

        logging.info(f"{self}: unsubscribe from topic {topic}")
        await self._on_unsubscribing(topic, **kwargs)
        await self._internal_unsubscribe(topic, **kwargs)
        await self._on_unsubscribed(topic, **kwargs)

    async def _on_subscribing(self, topic: str, **kwargs):
        """
        Callback called on subscribing

        :param topic:
        :return:
        """

    async def _on_subscribed(self, topic: str, **kwargs):
        """
        Callback called on subscribed

        :param topic:
        :return:
        """

    async def _on_unsubscribing(self, topic: Optional[str], **kwargs):
        """
        Callback called on unsubscribing

        :param topic:
        :return:
        """

    async def _on_unsubscribed(self, topic: Optional[str], **kwargs):
        """
        Callback called on unsubscribed

        :param topic:
        :return:
        """


import asyncio
from dataclasses import dataclass
from typing import List, Optional, override

from busline.client.client import EventBusClient
from busline.event.event import Event


@dataclass
class EventBusMultiClient(EventBusClient):

    clients: List[EventBusClient]

    @classmethod
    def from_client(cls, client: EventBusClient):
        return cls([client])

    @override
    async def connect(self):
        tasks = [client.connect() for client in self.clients]
        await asyncio.gather(*tasks)

    @override
    async def disconnect(self):
        tasks = [client.disconnect() for client in self.clients]
        await asyncio.gather(*tasks)

    @override
    async def publish(self, topic: str, event: Event, **kwargs):
        tasks = [client.publish(topic, event, **kwargs) for client in self.clients]
        await asyncio.gather(*tasks)

    @override
    async def subscribe(self, topic: str, **kwargs):
        tasks = [client.subscribe(topic, **kwargs) for client in self.clients]
        await asyncio.gather(*tasks)

    @override
    async def unsubscribe(self, topic: Optional[str] = None, **kwargs):
        tasks = [client.unsubscribe(topic, **kwargs) for client in self.clients]
        await asyncio.gather(*tasks)



from abc import ABC, abstractmethod
from busline.event.event import Event


class EventHandler(ABC):

    @abstractmethod
    async def handle(self, topic: str, event: Event):
        """
        Manage an event of a topic

        :param topic:
        :param event:
        :return:
        """


from typing import Callable, Optional, Self, override
from dataclasses import dataclass, field
from busline.client.publisher.publisher import Publisher
from busline.client.subscriber.event_handler.closure_event_handler import ClosureEventHandler
from busline.event.event import Event
from busline.client.pubsub_client import PubSubClient, PubSubClientBuilder, PubTopicSubClient
from busline.local.eventbus.eventbus import EventBus
from busline.local.eventbus.local_eventbus import LocalEventBus
from busline.local.publisher.local_publisher import LocalEventBusPublisher
from busline.local.subscriber.local_subscriber import LocalEventBusSubscriber


@dataclass
class LocalPubSubClient(PubSubClient):
    pass


@dataclass
class LocalPubTopicSubClient(PubTopicSubClient):
    pass


@dataclass
class LocalPubSubClientBuilder(PubSubClientBuilder):
    """
    Builder for a local pub/sub client.

    EventBus fed in init will be used to build publishers and subscribers

    Author: Nicola Ricciardi
    """

    eventbus: EventBus = field(default_factory=LocalEventBus)
    base_client: LocalPubSubClient = field(
        default_factory=lambda: LocalPubSubClient([], []),
        kw_only=True
    )

    def with_default_publisher(self) -> Self:
        """"
        LocalEventBusPublisher coupled with eventbus is used
        """

        self.base_client.publishers.append(
            LocalEventBusPublisher(eventbus=self.eventbus)
        )

        return self

    def with_closure_subscriber(self, closure: Callable[[str, Event], None]) -> Self:
        self.base_client.subscribers.append(
            LocalEventBusSubscriber(
                eventbus=self.eventbus,
                fallback_event_handler=ClosureEventHandler(closure)
            )
        )

        return self

    @override
    def build(self) -> LocalPubSubClient:
        return LocalPubSubClient.from_pubsub_client(self.base_client)


@dataclass
class LocalPubTopicSubClientBuilder(PubSubClientBuilder):
    """
    Builder for a local pub/sub client.

    EventBus fed in init will be used to build publishers and subscribers

    Author: Nicola Ricciardi
    """

    eventbus: EventBus = field(default_factory=LocalEventBus)
    base_client: LocalPubTopicSubClient = field(
        default_factory=lambda: LocalPubTopicSubClient([], []),
        kw_only=True
    )

    def with_default_publisher(self) -> Self:
        """"
        LocalEventBusPublisher coupled with eventbus is used
        """

        self.base_client.publishers.append(
            LocalEventBusPublisher(eventbus=self.eventbus)
        )

        return self

    def with_closure_subscriber(self, closure: Callable[[str, Event], None]) -> Self:
        self.base_client.subscribers.append(
            LocalEventBusSubscriber(
                eventbus=self.eventbus,
                fallback_event_handler=ClosureEventHandler(closure)
            )
        )

        return self

    @override
    def build(self) -> LocalPubTopicSubClient:
        return LocalPubTopicSubClient.from_pubsub_client(self.base_client)